﻿using System;
using System.Collections.Generic;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    internal struct ModelExEffectParameter
    {
        public String Name;
        public Object Value;
        public bool Constant;
    }

    class Material : Project.Resource
    {
        internal Effect Effect;
        public Texture Texture;
        internal List<ModelExEffectParameter> Parameters;
        internal EffectHandle ParameterBlock;

        Material(Project project, String name) :
            base(project, name)
        {
            Parameters = new List<ModelExEffectParameter>();
        }

        ~Material()
        {
        }

        public static Material Create(Project project, String name)
        {
            Material material = project.FindMaterial(name);
            return (material != null) ? material : new Material(project, name);
        }

        public override void Load(String folderName)
        {
        }

        public object GetParameter(int index)
        {
            if (index >= Parameters.Count)
            {
                return null;
            }

            return Parameters[index].Value;
        }

        public void SetParameters(ModelExEffectParameter[] parameters)
        {
            Parameters.AddRange(parameters);
            RefreshParameters();
        }

        public void RefreshParameters()
        {
            if (Effect == null)
            {
                return;
            }

            Effect.D3DXEffect.BeginParameterBlock();

            foreach (ModelExEffectParameter parameter in Parameters)
            {
                if (parameter.Constant)
                {
                    if (parameter.Value is Texture) // BaseTexture
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, ((Texture)parameter.Value).D3DXTexture);
                    }
                    else if (parameter.Value is bool)
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (bool)parameter.Value);
                    }
                    else if (parameter.Value is bool[])
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (bool[])parameter.Value);
                    }
                    else if (parameter.Value is ColorValue)
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (ColorValue)parameter.Value);
                    }
                    else if (parameter.Value is ColorValue[])
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (ColorValue[])parameter.Value);
                    }
                    else if (parameter.Value is GraphicsStream)
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (GraphicsStream)parameter.Value);
                    }
                    else if (parameter.Value is int)
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (int)parameter.Value);
                    }
                    else if (parameter.Value is int[])
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (int[])parameter.Value);
                    }
                    else if (parameter.Value is Matrix)
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (Matrix)parameter.Value);
                    }
                    else if (parameter.Value is Matrix[])
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (Matrix[])parameter.Value);
                    }
                    else if (parameter.Value is float)
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (float)parameter.Value);
                    }
                    else if (parameter.Value is float[])
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (float[])parameter.Value);
                    }
                    else if (parameter.Value is string)
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (string)parameter.Value);
                    }
                    else if (parameter.Value is Vector4)
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (Vector4)parameter.Value);
                    }
                    else if (parameter.Value is Vector4[])
                    {
                        Effect.D3DXEffect.SetValue(parameter.Name, (Vector4[])parameter.Value);
                    }
                }
            }

            ParameterBlock = Effect.D3DXEffect.EndParameterBlock();
        }
    }
}