﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    #region Enumerations

    public enum ControlType
    {
        None = 0,
        OrbitCamera = 1,
        FlyCamera = 2,
        MapCamera = 3,
        RotateObject = 4,
    }

    public enum ModelFormat
    {
        Binary = D3DX.XFileFormat.Binary,
        Text = D3DX.XFileFormat.Text,
        BinaryCompressed = D3DX.XFileFormat.Binary | D3DX.XFileFormat.Compressed,
        TextCompressed = D3DX.XFileFormat.Text | D3DX.XFileFormat.Compressed,
        LastUsed = 4
    }

    public enum ImageFormat
    {
        Bmp = 0,
        Jpeg = 1,
        Png = 2,
        LastUsed = 3
    }

    public enum CullType
    {
        None = D3DX.Cull.None,
        Clockwise = D3DX.Cull.Clockwise,
        CounterClockwise = D3DX.Cull.CounterClockwise
    }

    public enum RenderFlags
    {
        None = 0x00,
        Lighting = 0x01,
        Textures = 0x02,
        TextureFiltering = 0x04,
        WireFrame = 0x08,
        Scaling = 0x10,
    }

    public enum DebugType
    {
        None = 0,
        Debug = 1,
        Textures = 2,
        Materials = 3,
        Effects = 4,
        Meshes = 5,
    }

    #endregion

    class Physical : Project.Resource
    {
        internal Mesh Mesh;
        Quaternion RotationValue;
        public ArcBall ArcBall;
        public Vector3 Position;

        public Quaternion Rotation
        {
            get { return RotationValue; }
            set
            {
                RotationValue = value;
                if (ArcBall != null)
                {
                    ArcBall.Rotation = RotationValue;
                }
            }
        }

        Physical(Project project, String name) :
            base(project, name)
        {
            RotationValue = Quaternion.Identity;
            Position = new Vector3(0.0f, 0.0f, 0.0f);
            Project = project;
        }

        public static Physical Create(Project project, String name)
        {
            Physical newObject = project.FindObject(name);
            return (newObject != null) ? newObject : new Physical(project, name);
        }

        public override void Load(string folderName)
        {
        }

        public void Render(RenderFlags renderFlags, CullType cullType, float fscale)
        {
            if (ArcBall != null)
            {
                RotationValue = ArcBall.Rotation;
            }

            Vector3 scale = Mesh.Scale;
            if ((renderFlags & RenderFlags.Scaling) == 0)
            {
                scale = new Vector3(fscale, fscale, fscale);
            }

            Matrix World =
            Matrix.Translation(Position - Mesh.Centre) *
            Matrix.Scaling(scale) *
            Matrix.RotationQuaternion(RotationValue);

            if (Mesh != null)
            {
                Mesh.Render(World, renderFlags, cullType);
            }
        }
    }
}
