﻿using System;
using System.IO;
using System.Reflection;
using System.Collections.Generic;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    partial class Project
    {
        double[] FloatsToDoubles(float[] floats)
        {
            if (floats == null)
            {
                return null;
            }

            double[] doubles = new double[floats.Length];
            for (int i = 0; i < floats.Length; i++)
            {
                doubles[i] = (double)floats[i];
            }

            return doubles;
        }

        float[] DoublesToFloats(double[] doubles)
        {
            if (doubles == null)
            {
                return null;
            }

            float[] floats = new float[doubles.Length];
            for (int i = 0; i < doubles.Length; i++)
            {
                floats[i] = (float)doubles[i];
            }

            return floats;
        }

        float[] FlipYZAxis(float[] floats)
        {
            float[] flippedFloats = new float[floats.Length];
            for (int i = 0; i < flippedFloats.Length; i++)
            {
                if (i % 3 == 0)
                {
                    flippedFloats[i] = floats[i];
                }
                if (i % 3 == 1)
                {
                    flippedFloats[i] = floats[i + 1];
                }
                if (i % 3 == 2)
                {
                    flippedFloats[i] = floats[i - 1];
                }
            }
            return flippedFloats;
        }

        float[] FlipTextureCoords(float[] floats)
        {
            float[] flippedFloats = new float[floats.Length];
            for (int i = 0; i < flippedFloats.Length; i++)
            {
                if (i % 2 == 0)
                {
                    flippedFloats[i] = floats[i];
                }
                if (i % 2 == 1)
                {
                    flippedFloats[i] = (floats[i] * -1.0f) + 1.0f;
                }
            }
            return flippedFloats;
        }

        String UsageToSemantic(DeclarationUsage usage)
        {
            switch (usage)
            {
                case DeclarationUsage.Position:
                    return "VERTEX";
                case DeclarationUsage.BlendWeight:
                    return "WEIGHT";
                case DeclarationUsage.BlendIndices:
                    return "JOINT";
                case DeclarationUsage.Normal:
                    return "NORMAL";
                case DeclarationUsage.PointSize:
                    return "";
                case DeclarationUsage.TextureCoordinate:
                    return "TEXCOORD";
                case DeclarationUsage.Tangent:
                    return "TANGENT";
                case DeclarationUsage.BiNormal:
                    return "BINORMAL";
                case DeclarationUsage.TessellateFactor:
                    return "";
                case DeclarationUsage.PositionTransformed:
                    return "";
                case DeclarationUsage.Color:
                    return "COLOR";
                case DeclarationUsage.Fog:
                    return "";
                case DeclarationUsage.Depth:
                    return "";
                case DeclarationUsage.Sample:
                    return "";
                default:
                    return null;
            }
        }

        Collada141.common_color_or_texture_type ColorToCollada(float[] colorIn)
        {
            Collada141.common_color_or_texture_typeColor colladaColor = new Collada141.common_color_or_texture_typeColor();
            colladaColor.Values = FloatsToDoubles(colorIn);
            Collada141.common_color_or_texture_type colladaColorOrTexture = new Collada141.common_color_or_texture_type();
            colladaColorOrTexture.Item = colladaColor;
            return colladaColorOrTexture;
        }

        Collada141.common_color_or_texture_type TextureToCollada(String textureIn, String texCoordIn)
        {
            Collada141.common_color_or_texture_typeTexture colladaTexture = new Collada141.common_color_or_texture_typeTexture();
            colladaTexture.texture = textureIn;
            colladaTexture.texcoord = texCoordIn;
            Collada141.common_color_or_texture_type colladaColorOrTexture = new Collada141.common_color_or_texture_type();
            colladaColorOrTexture.Item = colladaTexture;
            return colladaColorOrTexture;
        }

        Collada141.common_float_or_param_type FloatToCollada(float floatIn)
        {
            Collada141.common_float_or_param_typeFloat colladaFloat = new Collada141.common_float_or_param_typeFloat();
            colladaFloat.Value = (float) floatIn;
            Collada141.common_float_or_param_type colladaFloatOrParam = new Collada141.common_float_or_param_type();
            colladaFloatOrParam.Item = colladaFloat;
            return colladaFloatOrParam;
        }

        Collada141.source FloatSourceToCollada(FloatSourceBase sourceIn)
        {
            Collada141.float_array floatArray = new Collada141.float_array();
            floatArray.id = sourceIn.Name + "-array";
            float[] floats = sourceIn.Floats;
            if (sourceIn.Usage == DeclarationUsage.Position)
            {
                floats = FlipYZAxis(floats);
            }
            else if (sourceIn.Usage == DeclarationUsage.TextureCoordinate)
            {
                floats = FlipTextureCoords(floats);
            }
            floatArray.Values = FloatsToDoubles(floats);
            floatArray.count = (ulong)sourceIn.Floats.Length;
            Collada141.param[] parameters = new Collada141.param[sourceIn.Stride];
            for (int i = 0; i < sourceIn.Stride; i++)
            {
                Collada141.param parameter = new Collada141.param();
                parameter.name = sourceIn.ParameterNames[i];
                parameter.type = "float";
                parameters[i] = parameter;
            }
            Collada141.accessor accessor = new Collada141.accessor();
            accessor.param = parameters;
            accessor.source = "#" + sourceIn.Name + "-array";
            accessor.stride = (ulong)sourceIn.Stride;
            accessor.count = (ulong)sourceIn.Count;
            Collada141.sourceTechnique_common techniqueCommon = new Collada141.sourceTechnique_common();
            techniqueCommon.accessor = accessor;
            Collada141.source source = new Collada141.source();
            source.id = sourceIn.Name;
            source.Item = floatArray;
            source.technique_common = techniqueCommon;
            return source;
        }

        public void LoadColladaFile(String strFileName)
        {
            Collada141.COLLADA colladaDoc = Collada141.COLLADA.Load(strFileName);
        }

        public void SaveColladaFile(String strFileName)
        {
            Assembly executingAssembly = Assembly.GetExecutingAssembly();
            Stream stream = executingAssembly.GetManifestResourceStream(GetType(), "Collada141.dae");
            Collada141.COLLADA colladaDoc = Collada141.COLLADA.Load(stream);

            Collada141.assetContributor[] contributers = new Collada141.assetContributor[1];
            Collada141.assetContributor contributer = new Collada141.assetContributor();
            contributer.author = "ModelEx User";
            contributer.authoring_tool = "ModelEx 4.0";
            contributers[0] = contributer;
            colladaDoc.asset.contributor = contributers;

            DateTime now = DateTime.SpecifyKind(DateTime.Now, DateTimeKind.Unspecified);
            colladaDoc.asset.created = now;
            colladaDoc.asset.modified = now;

            Collada141.assetUnit unit = new Collada141.assetUnit();
            unit.name = "meter";
            unit.meter = 1.0f;
            colladaDoc.asset.unit = unit;

            colladaDoc.asset.up_axis = Collada141.UpAxisType.Z_UP;

            List<Collada141.image> images = new List<Collada141.image>();
            List<Collada141.effect> effects = new List<Collada141.effect>();
            List<Collada141.material> materials = new List<Collada141.material>();
            List<Collada141.geometry> geometries = new List<Collada141.geometry>();
            List<Collada141.visual_scene> visualScenes = new List<Collada141.visual_scene>();

            foreach (Resource Reference in TextureReferences)
            {
                Texture textureIn = (Texture)Reference;
                if (textureIn.Name != "")
                {
                    Collada141.image imageOut = new Collada141.image();
                    imageOut.id = "image-" + textureIn.Name;
                    imageOut.name = "image-" + textureIn.Name;
                    imageOut.Item = Path.ChangeExtension(textureIn.Name, "dds");
                    images.Add(imageOut);

                    String strTextureFilePath = Path.GetDirectoryName(strFileName);
                    strTextureFilePath = Path.Combine(strTextureFilePath, (String)imageOut.Item);
                    textureIn.Save(strTextureFilePath);
                }
            }

            foreach (Resource Reference in MaterialReferences)
            {
                Material materialIn = (Material)Reference;
                if (materialIn.Name != "")
                {
                    Collada141.common_newparam_type[] parameters = null;

                    bool useTexture = (materialIn.Texture != null && materialIn.Texture.Name != "");

                    if (useTexture)
                    {
                        parameters = new Collada141.common_newparam_type[2];

                        Collada141.common_newparam_type surfaceNewParam = new Collada141.common_newparam_type();
                        surfaceNewParam.sid = "xSurface0";
                        surfaceNewParam.ItemElementName = Collada141.ItemChoiceType.surface;
                        Collada141.fx_surface_init_from_common initFrom = new Collada141.fx_surface_init_from_common();
                        initFrom.Value = "image-" + materialIn.Texture.Name;
                        Collada141.fx_surface_common surface = new Collada141.fx_surface_common();
                        surface.type = Collada141.fx_surface_type_enum.Item2D;
                        Collada141.fx_surface_init_from_common[] initFroms = new Collada141.fx_surface_init_from_common[1];
                        initFroms[0] = initFrom;
                        surface.init_from = initFroms;
                        surfaceNewParam.Item = surface;
                        parameters[0] = surfaceNewParam;

                        Collada141.common_newparam_type textureNewParam = new Collada141.common_newparam_type();
                        textureNewParam.sid = "xTexture0";
                        textureNewParam.ItemElementName = Collada141.ItemChoiceType.sampler2D;
                        Collada141.fx_sampler2D_common sampler2D = new Collada141.fx_sampler2D_common();
                        sampler2D.source = surfaceNewParam.sid;
                        textureNewParam.Item = sampler2D;
                        parameters[1] = textureNewParam;
                    }

                    Collada141.effectFx_profile_abstractProfile_COMMONTechniquePhong techniquePhong =
                        new Collada141.effectFx_profile_abstractProfile_COMMONTechniquePhong();

                    float[] emissionFloats = { 0.0f, 0.0f, 0.0f, 1.0f };
                    techniquePhong.emission = ColorToCollada(emissionFloats);

                    float[] ambientFloats = { 0.0f, 0.0f, 0.0f, 1.0f };
                    techniquePhong.ambient = ColorToCollada(ambientFloats);

                    if (useTexture)
                    {
                        techniquePhong.diffuse = TextureToCollada("xTexture0", "UVTex");
                    }
                    else
                    {
                        float[] diffuseFloats = { 1.0f, 1.0f, 1.0f, 1.0f };
                        techniquePhong.diffuse = ColorToCollada(diffuseFloats);
                    }

                    float[] specularFloats = { 0.0f, 0.0f, 0.0f, 1.0f };
                    techniquePhong.specular = ColorToCollada(specularFloats);

                    techniquePhong.shininess = FloatToCollada(50.0f);

                    techniquePhong.index_of_refraction = FloatToCollada(1.0f);

                    Collada141.effectFx_profile_abstractProfile_COMMONTechnique techniqueCommon =
                        new Collada141.effectFx_profile_abstractProfile_COMMONTechnique();
                    techniqueCommon.sid = "common";
                    techniqueCommon.Item = techniquePhong;

                    Collada141.effectFx_profile_abstractProfile_COMMON profileCommon =
                        new Collada141.effectFx_profile_abstractProfile_COMMON();
                    profileCommon.Items = parameters;
                    profileCommon.technique = techniqueCommon;

                    Collada141.effectFx_profile_abstractProfile_COMMON[] profiles = new Collada141.effectFx_profile_abstractProfile_COMMON[1];
                    profiles[0] = profileCommon;

                    Collada141.effect effectOut = new Collada141.effect();
                    effectOut.id = "effect-" + materialIn.Name;
                    effectOut.Items = profiles;

                    effects.Add(effectOut);

                    Collada141.instance_effect instanceEffect = new Collada141.instance_effect();
                    instanceEffect.url = "#effect-" + materialIn.Name;

                    Collada141.material materialOut = new Collada141.material();
                    materialOut.id = "material-" + materialIn.Name;
                    materialOut.name = "material-" + materialIn.Name;
                    materialOut.instance_effect = instanceEffect;

                    materials.Add(materialOut);
                }
            }

            foreach (Resource Reference in MeshReferences)
            {
                Mesh meshIn = (Mesh)Reference;
                if (meshIn.Name != "")
                {
                    Collada141.source[] sources = new Collada141.source[4];

                    sources[0] = FloatSourceToCollada(meshIn.Positions);

                    sources[1] = FloatSourceToCollada(meshIn.Normals);

                    sources[2] = FloatSourceToCollada(meshIn.Colors);

                    sources[3] = FloatSourceToCollada(meshIn.TextureCoords);

                    Collada141.InputLocal[] verticesInputs = new Collada141.InputLocal[1];

                    Collada141.InputLocal positionInput = new Collada141.InputLocal();
                    positionInput.semantic = "POSITION";
                    positionInput.source = "#" + sources[0].id;
                    verticesInputs[0] = positionInput;
                    Collada141.vertices vertices = new Collada141.vertices();
                    vertices.id = meshIn.Name + "-vertices";
                    vertices.input = verticesInputs;

                    Object[] primitives = new Object[meshIn.Triangles.Length];
                    for (int i = 0; i < meshIn.Triangles.Length; i++)
                    {
                        if (meshIn.Triangles[i] != null &&
                            meshIn.Triangles[i].Material != null &&
                            meshIn.Triangles[i].IndexCount > 0)
                        {
                            Triangles.SourceMapping[] mappings = meshIn.Triangles[i].SourceMappings.ToArray();
                            int numMappings = mappings.Length;

                            List<Collada141.InputLocalOffset> trianglesInputs = new List<Collada141.InputLocalOffset>();

                            for (int m = 0; m < numMappings; m++)
                            {
                                foreach (Source source in mappings[m].Sources)
                                {
                                    Collada141.InputLocalOffset input = new Collada141.InputLocalOffset();
                                    input.semantic = UsageToSemantic(source.Usage);
                                    input.source = "#" + source.Name;
                                    input.offset = (ulong)m;
                                    if (source.Usage == DeclarationUsage.TextureCoordinate)
                                    {
                                        input.set = (ulong)0;
                                        input.setSpecified = true;
                                    }
                                    trianglesInputs.Add(input);
                                }
                            }

                            String p = "";
                            for (int v = 0; v < meshIn.Triangles[i].IndexCount; v++)
                            {
                                for (int m = 0; m < numMappings; m++)
                                {
                                    p += mappings[m].Indices[v].ToString() + " ";
                                }
                            }

                            Collada141.triangles triangles = new Collada141.triangles();
                            triangles.material = "material-" + meshIn.Triangles[i].Material.Name;
                            triangles.count = (ulong)meshIn.Triangles[i].IndexCount / 3;
                            triangles.p = p;
                            triangles.input = trianglesInputs.ToArray();
                            primitives[i] = triangles;
                        }
                    }

                    Collada141.mesh mesh = new Collada141.mesh();
                    mesh.source = sources;
                    mesh.vertices = vertices;
                    mesh.Items = primitives;

                    Collada141.geometry geometryOut = new Collada141.geometry();
                    geometryOut.id = "geometry-" + meshIn.Name;
                    geometryOut.Item = mesh;

                    geometries.Add(geometryOut);
                }
            }

            List<Collada141.node> nodes = new List<Collada141.node>();
            foreach (Resource Reference in PhysicalReferences)
            {
                Physical physicalIn = (Physical)Reference;
                if (physicalIn.Name != "")
                {
                    Collada141.instance_material[] instanceMaterials = new Collada141.instance_material[physicalIn.Mesh.Materials.Length];
                    for (int i = 0; i < physicalIn.Mesh.Materials.Length; i++)
                    {
                        Collada141.instance_materialBind_vertex_input[] bindVertexInputs = new Collada141.instance_materialBind_vertex_input[1];
                        Collada141.instance_materialBind_vertex_input bindVertexInput = new Collada141.instance_materialBind_vertex_input();
                        bindVertexInput.semantic = "UVTex";
                        bindVertexInput.input_semantic = "TEXCOORD";
                        bindVertexInput.input_set = (ulong)0;
                        bindVertexInput.input_setSpecified = true;
                        bindVertexInputs[0] = bindVertexInput;

                        Collada141.instance_material instanceMaterial = new Collada141.instance_material();
                        instanceMaterial.symbol = "material-" + physicalIn.Mesh.Materials[i].Name;
                        instanceMaterial.target = "#material-" + physicalIn.Mesh.Materials[i].Name;
                        instanceMaterial.bind_vertex_input = bindVertexInputs;
                        instanceMaterials[i] = instanceMaterial;
                    }

                    Collada141.bind_material bindMaterial = new Collada141.bind_material();
                    bindMaterial.technique_common = instanceMaterials;

                    Collada141.instance_geometry[] instanceGeometries = new Collada141.instance_geometry[1];
                    Collada141.instance_geometry instanceGeometry = new Collada141.instance_geometry();
                    instanceGeometry.url = "#geometry-" + physicalIn.Name;
                    instanceGeometry.bind_material = bindMaterial;
                    instanceGeometries[0] = instanceGeometry;

                    Collada141.node nodeOut = new Collada141.node();
                    nodeOut.id = physicalIn.Name;
                    nodeOut.name = physicalIn.Name;
                    nodeOut.type = Collada141.NodeType.NODE;
                    nodeOut.instance_geometry = instanceGeometries;
                    nodes.Add(nodeOut);
                }
            }

            Collada141.visual_scene visualSceneOut = new Collada141.visual_scene();
            visualSceneOut.id = "scene";
            visualSceneOut.name = "scene";
            visualSceneOut.node = nodes.ToArray();
            visualScenes.Add(visualSceneOut);

            Collada141.library_images libraryImages = new Collada141.library_images();
            Collada141.library_effects libraryEffects = new Collada141.library_effects();
            Collada141.library_materials libraryMaterials = new Collada141.library_materials();
            Collada141.library_geometries libraryGeometries = new Collada141.library_geometries();
            Collada141.library_visual_scenes libraryVisualScenes = new Collada141.library_visual_scenes();

            libraryImages.image = images.ToArray();
            libraryEffects.effect = effects.ToArray();
            libraryMaterials.material = materials.ToArray();
            libraryGeometries.geometry = geometries.ToArray();
            libraryVisualScenes.visual_scene = visualScenes.ToArray();

            List<Object> libraries = new List<Object>();

            libraries.Add(libraryImages);
            libraries.Add(libraryEffects);
            libraries.Add(libraryMaterials);
            libraries.Add(libraryGeometries);
            libraries.Add(libraryVisualScenes);

            colladaDoc.Items = libraries.ToArray();

            Collada141.InstanceWithExtra instanceVisualScene = new Collada141.InstanceWithExtra();
            instanceVisualScene.url = "#scene";
            Collada141.COLLADAScene scene = new Collada141.COLLADAScene();
            scene.instance_visual_scene = instanceVisualScene;

            colladaDoc.scene = scene;

            colladaDoc.Save(strFileName);
        }
    }
}
