﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;
using BenLincoln.TheLostWorlds.CDTextures;
using CDTextureFile = BenLincoln.TheLostWorlds.CDTextures.TextureFile;
using SR1PCTextureFile = BenLincoln.TheLostWorlds.CDTextures.SoulReaverPCTextureFile;
using SR1PSTextureFile = BenLincoln.TheLostWorlds.CDTextures.SoulReaverPlaystationTextureFile;
using SR1DCTextureFile = BenLincoln.TheLostWorlds.CDTextures.SoulReaverDreamcastTextureFile;
using SR2PCTextureFile = BenLincoln.TheLostWorlds.CDTextures.SoulReaver2PCVRMTextureFile;
using VertexPNT = Microsoft.DirectX.Direct3D.CustomVertex.PositionNormalTextured;
using VertexPCT = Microsoft.DirectX.Direct3D.CustomVertex.PositionColoredTextured;

namespace ModelEx
{
    partial class Project
    {
        struct VertexPNCT
        {
            public const VertexFormats Format = VertexFormats.Position | VertexFormats.Normal | VertexFormats.Diffuse | VertexFormats.Texture1;

            public float X;
            public float Y;
            public float Z;
            public float Nx;
            public float Ny;
            public float Nz;
            public int Color;
            public float Tu;
            public float Tv;

            public VertexPNCT(Vector3 pos, Vector3 nor, int c, float u, float v)
            {
                X = pos.X;
                Y = pos.Y;
                Z = pos.Z;

                Nx = nor.X;
                Ny = nor.Y;
                Nz = nor.Z;

                Color = c;

                Tu = u;
                Tv = v;
            }

            public VertexPNCT(float xvalue, float yvalue, float zvalue, float nxvalue, float nyvalue, float nzvalue, int c, float u, float v)
            {
                X = xvalue;
                Y = yvalue;
                Z = zvalue;

                Nx = nxvalue;
                Ny = nyvalue;
                Nz = nzvalue;

                Color = c;

                Tu = u;
                Tv = v;
            }

            public Vector3 Normal
            {
                get { return new Vector3(Nx, Ny, Nz); }
                set { Nx = value.X; Ny = value.Y; Nz = value.Z; }
            }
            public Vector3 Position
            {
                get { return new Vector3(X, Y, Z); }
                set { X = value.X; Y = value.Y; Z = value.Z; }
            }
            public static int StrideSize
            {
                get { return 36; }
            }

            public static VertexElement[] Elements = new VertexElement[]
		    {
			    new VertexElement(0, 0, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Position, 0),
		        new VertexElement(0, 12, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Normal, 0),
		        new VertexElement(0, 24, DeclarationType.Color, DeclarationMethod.Default, DeclarationUsage.Color, 0),
		        new VertexElement(0, 28, DeclarationType.Float2, DeclarationMethod.Default, DeclarationUsage.TextureCoordinate, 0),
			    VertexElement.VertexDeclarationEnd
    		};
        }

        SRFile SRFile;
        bool isSR2File;

        public String TextureArchives
        {
            get
            {
                if (SRFile == null)
                {
                    return null;
                }

                switch (SRFile.m_ePlatform)
                {
                    case Platform.PC:
                        return
                            "Soul Reaver 2 VRM Files (*.vrm)|*.vrm|" +
                            //"Soul Reaver Monolithic Files (*.big)|*.big|" +
                            "All Files (*.*)|*.*";
                    case Platform.PSX:
                        return
                            "Soul Reaver CRM Files (*.crm)|*.crm|" +
                            //"Soul Reaver Archive Files (*.dat)|*.dat|" +
                            "All Files (*.*)|*.*";
                    case Platform.Dreamcast:
                        return
                            "Soul Reaver Archive Files (*.vq)|*.vq|" +
                            "All Files (*.*)|*.*";
                    default:
                        return null;
                }
            }
        }

        public String TextureArchiveExt
        {
            get
            {
                switch (SRFile.m_ePlatform)
                {
                    case Platform.PC: return "vrm"; //"big";
                    case Platform.PSX: return "crm";
                    case Platform.Dreamcast: return "vq";
                    default: return null;
                }
            }
        }

        public void LoadSoulReaverFile(String strFileName)
        {
            SRFile = new SR2File(strFileName);
            isSR2File = true;

            String objectName = Path.GetFileNameWithoutExtension(strFileName);
            String meshName = objectName + "-0";

            Material[] materials = new Material[SRFile.m_axModels[0].MaterialCount];
            for (int m = 0; m < SRFile.m_axModels[0].MaterialCount; m++)
            {
                Material material = Material.Create(this, meshName + "-" + m.ToString());
                material.Effect = DefaultEffect;
                material.Texture = Texture.Create(this, GetTextureName(SRFile.m_axModels[0], m));
                ModelExEffectParameter[] parameters = CreateEffectParameters(SRFile.m_axModels[0], m);
                material.SetParameters(parameters);
                materials[m] = material;
            }

            int vertexCount = (int)SRFile.m_axModels[0].Vertices.Length;
            int positionCount = (int)SRFile.m_axModels[0].Positions.Length;
            int normalCount = (int)SRFile.m_axModels[0].Normals.Length;
            int colorCount = (int)SRFile.m_axModels[0].Colours.Length;
            int textureCoordCount = (int)SRFile.m_axModels[0].UVs.Length;
            Vector3Source positions = Vector3Source.Create(meshName + "-positions", DeclarationUsage.Position, positionCount, "X", "Y", "Z");
            Vector3Source normals = Vector3Source.Create(meshName + "-normals", DeclarationUsage.Normal, normalCount, "X", "Y", "Z");
            ColorValueSource colors = ColorValueSource.Create(meshName + "-colors", DeclarationUsage.Color, colorCount, "R", "G", "B");
            Vector2Source textureCoords = Vector2Source.Create(meshName + "-textureCoords", DeclarationUsage.TextureCoordinate, textureCoordCount, "U", "V");
            for (int v = 0; v < vertexCount; v++)
            {
                ExVertex vert = SRFile.m_axModels[0].Vertices[v];

                Vector3 position = new Vector3();
                position.X = 0.001f * SRFile.m_axModels[0].Positions[vert.positionID].worldPos.x; // 100.0f;
                position.Y = 0.001f * SRFile.m_axModels[0].Positions[vert.positionID].worldPos.z; // 100.0f;
                position.Z = 0.001f * SRFile.m_axModels[0].Positions[vert.positionID].worldPos.y; // 100.0f;
                positions[vert.positionID] = position;

                Vector3 normal = new Vector3();
                normal.X = (float)SRFile.m_axModels[0].Normals[vert.normalID].x;
                normal.Y = (float)SRFile.m_axModels[0].Normals[vert.normalID].z;
                normal.Z = (float)SRFile.m_axModels[0].Normals[vert.normalID].y;
                normal.Normalize();
                normals[vert.normalID] = normal;

                int color = (int)SRFile.m_axModels[0].Colours[vert.colourID];
                colors[vert.colourID] = ColorValue.FromArgb(color);

                Vector2 textureCoordinate = new Vector2();
                textureCoordinate.X = SRFile.m_axModels[0].UVs[vert.UVID].u;
                textureCoordinate.Y = SRFile.m_axModels[0].UVs[vert.UVID].v;
                textureCoords[vert.UVID] = textureCoordinate;
            }

            Triangles[] triangles = new Triangles[SRFile.m_axModels[0].MaterialCount];
            for (int m = 0; m < SRFile.m_axModels[0].MaterialCount; m++)
            {
                triangles[m] = Triangles.Create(this, meshName + "-" + m.ToString());
                triangles[m].Material = FindMaterial(meshName + "-" + m.ToString());

                Int32 indexCount = 0;
                for (UInt32 i = 0; i < vertexCount; i++)
                {
                    if (SRFile.m_axModels[0].Polygons[i / 3].material.ID != m ||
                        !SRFile.m_axModels[0].Polygons[i / 3].material.visible)
                    {
                        continue;
                    }

                    indexCount++;
                }

                if (indexCount == 0)
                {
                    continue;
                }

                VertexPNCT[] verticesPNCT = new VertexPNCT[indexCount];
                Int32 vertex = 0;
                for (int v = 0; v < vertexCount; v++)
                {
                    if (SRFile.m_axModels[0].Polygons[v / 3].material.ID != m)
                    {
                        continue;
                    }

                    ExVertex vert = SRFile.m_axModels[0].Vertices[v];

                    verticesPNCT[vertex] = new VertexPNCT(
                        positions[vert.positionID].X,
                        positions[vert.positionID].Y,
                        positions[vert.positionID].Z,
                        normals[vert.normalID].X,
                        normals[vert.normalID].Y,
                        normals[vert.normalID].Z,
                        colors[vert.colourID].ToArgb(),
                        textureCoords[vert.UVID].X,
                        textureCoords[vert.UVID].Y
                    );

                    vertex++;
                }
                VertexBuffer vertexBuffer = new VertexBuffer(typeof(VertexPNCT), verticesPNCT.Length, this.Device, Usage.None, VertexPNCT.Format, Pool.Managed);
                vertexBuffer.SetData(verticesPNCT, 0, LockFlags.None);
                triangles[m].VertexBuffer = vertexBuffer;
                triangles[m].VertexCount = verticesPNCT.Length;
                triangles[m].VertexFormat = VertexPNCT.Format;
                triangles[m].Stride = VertexPNCT.StrideSize;

                Int32[] indices = new Int32[indexCount];
                Int32[] positionIDs = new Int32[indexCount];
                Int32[] normalIDs = new Int32[indexCount];
                Int32 index = 0;
                for (int i = 0; i < vertexCount; i++)
                {
                    if (SRFile.m_axModels[0].Polygons[i / 3].material.ID != m)
                    {
                        continue;
                    }

                    ExVertex vert = SRFile.m_axModels[0].Vertices[i];

                    indices[index] = index;
                    positionIDs[index] = vert.positionID;
                    normalIDs[index] = vert.normalID;
                    index++;
                }
                IndexBuffer indexBuffer = new IndexBuffer(typeof(Int32), indices.Length, this.Device, Usage.None, Pool.Managed);
                indexBuffer.SetData(indices, 0, LockFlags.None);
                triangles[m].IndexBuffer = indexBuffer;
                triangles[m].IndexCount = indices.Length;
                Source[] positionSources = { positions, colors, textureCoords };
                Source[] normalSources = { normals };
                triangles[m].AddSourceMapping(positionSources, positionIDs);
                triangles[m].AddSourceMapping(normalSources, normalIDs);
            }

            Mesh mesh = Mesh.Create(this, meshName);
            mesh.Positions = positions;
            mesh.Normals = normals;
            mesh.Colors = colors;
            mesh.TextureCoords = textureCoords;
            mesh.Materials = materials;
            mesh.Triangles = triangles;
            mesh.ComputeBoundingSphere();

            PhysicalReferences.Clear();
            Physical mainObject = Physical.Create(this, meshName);
            mainObject.Mesh = mesh;

            _Name = meshName;

            String textureName = Path.ChangeExtension(strFileName, "vrm");
            LoadSoulReaver2Textures(textureName);
        }

        ModelExEffectParameter[] CreateEffectParameters(SRModel srModel, int iMaterial)
        {
            ModelExEffectParameter[] parameters = new ModelExEffectParameter[10];

            parameters[0].Name = "iAmbientMaterialSource";
            parameters[0].Constant = true;
            parameters[0].Value = (int)ColorSource.Material;
            parameters[1].Name = "iDiffuseMaterialSource";
            parameters[1].Constant = true;
            parameters[1].Value = (int)ColorSource.Color1;
            parameters[2].Name = "bColorVertex";
            parameters[2].Constant = true;
            parameters[2].Value = true; // false for SR1 objects

            if (SRFile.m_eFileType == FileType.Area)
            {
                parameters[0].Value = (int)ColorSource.Color1;
                parameters[2].Value = true;
            }

            ColorValue ambientValue;
            ColorValue diffuseValue;
            if (SRFile.m_eFileType == FileType.Object &&
                srModel.Materials[iMaterial].textureUsed)
            {
                ambientValue = ColorValue.FromArgb(unchecked((int)0xFF404040));
                diffuseValue = ColorValue.FromArgb(unchecked((int)0xFFFFFFFF));
            }
            else
            {
                ambientValue = ColorValue.FromArgb(unchecked((int)0x00000000));
                diffuseValue = ColorValue.FromArgb((Int32)(srModel.Materials[iMaterial].colour));
            }

            float[] ambient = { ambientValue.Red, ambientValue.Green, ambientValue.Blue, ambientValue.Alpha };
            float[] diffuse = { diffuseValue.Red, diffuseValue.Green, diffuseValue.Blue, diffuseValue.Alpha };

            parameters[3].Name = "xMaterialAmbient";
            parameters[3].Constant = true;
            parameters[3].Value = ambient;
            parameters[4].Name = "xMaterialDiffuse";
            parameters[4].Constant = true;
            parameters[4].Value = diffuse;
            parameters[5].Name = "xTexture0";
            parameters[5].Constant = true;
            parameters[5].Value = FindTexture(GetTextureName(srModel, iMaterial));
            parameters[6].Name = "bLighting";
            parameters[6].Constant = false;
            parameters[6].Value = true;
            parameters[7].Name = "iColorOperation";
            parameters[7].Constant = false;
            parameters[7].Value = TextureOperation.Modulate;
            parameters[8].Name = "iTextureFilter";
            parameters[8].Constant = false;
            parameters[8].Value = TextureFilter.Linear;
            parameters[9].Name = "iFillMode";
            parameters[9].Constant = false;
            parameters[9].Value = FillMode.Solid;

            if (SRFile.m_eFileType == FileType.Area)
            {
                parameters[6].Value = false;

                if (!isSR2File)
                {
                    parameters[7].Value = TextureOperation.Modulate4X;
                }
            }

            return parameters;

            // ColorSource.Color1 = Vertex.DiffuseColor
            // ColorSource.Color2 = Vertex.SpecularColor
            // AmbientIntensity = Device.Ambient
            // AmbientColor = AmbientSource
            // DiffuseIntensity = Device.Lights[0].Diffuse
            // DiffuseColor = DiffuseSource
            // SpecularIntensity = Device.Lights[0].Specular
            // SpecularColor = SpecularSource

            // Ambient = Ca*[Ga + sum(Atteni*Spoti*Lai)]
            // Diffuse = sum[Cd*Ld*(N.Ldir)*Attenuation*Spot]
            // Specular = Cs*sum[Ls*(N·H)P*Attenuation*Spot]

            // I = (Ambient) + (Diffuse) + (Specular)
        }

        protected String GetTextureName(SRModel srModel, int iMaterial)
        {
            ExMaterial xMaterial = srModel.Materials[iMaterial];
            String textureName = "";
            if (xMaterial.textureUsed)
            {
                if (srModel is SR1Model)
                {
                    if (srModel.Platform == Platform.PSX)
                    {
                        textureName =
                            srModel.Name.TrimEnd(new char[] { '_' }).ToLower() + "-" +
                            xMaterial.textureID.ToString("0000");
                    }
                    else
                    {
                        textureName =
                            "Texture-" +
                            xMaterial.textureID.ToString("00000");
                    }
                }
                else if (srModel is SR2Model)
                {
                    textureName =
                        srModel.Name.TrimEnd(new char[] { '_' }).ToLower() + "-" +
                        xMaterial.textureID.ToString("0000");
                }
            }

            return textureName;
        }

        public void LoadSoulReaver2Textures(String strFileName)
        {
            SoulReaver2PCVRMTextureFile textureFile = new SoulReaver2PCVRMTextureFile(strFileName);
            String objectName = Path.GetFileNameWithoutExtension(strFileName);

            for (int t = 0; t < textureFile.TextureCount; t++)
            {
                String textureName =
                    objectName.TrimEnd(new char[] { '_' }).ToLower() + "-" +
                    textureFile.TextureDefinitions[t].Flags1.ToString("0000");

                Texture texture = FindTexture(textureName);
                if (texture != null)
                {
                    texture.D3DXTexture = textureFile.GetTexture(this.Device, t);
                }
            }

            for (int m = 0; m < MaterialReferences.Count; m++)
            {
                Material material = (Material)MaterialReferences[m];
                material.RefreshParameters();
            }
        }
    }
}
