﻿using System;
using System.Drawing;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    class ModelExQuad
    {
        protected Device Device;
        public Rectangle Rect;
        public float U1 = 0.0f, V1 = 0.0f;
        public float U2 = 1.0f, V2 = 1.0f;
        public UInt32 Colour = 0xFFFFFFFF;
        public UInt32 BorderColour = 0x00000000;
        public bool Enabled;

        public ModelExQuad(Device device, Rectangle rect, UInt32 colour, UInt32 borderColour)
        {
            Device = device;
            Rect = rect;
            Colour = colour;
            BorderColour = borderColour;
            Enabled = true;
        }

        public ModelExQuad(Device device, Rectangle rect, float u1, float v1, float u2, float v2, UInt32 colour, UInt32 borderColour)
        {
            Device = device;
            Rect = rect;
            U1 = u1;
            V1 = v1;
            U2 = u2;
            V2 = v2;
            Colour = colour;
            BorderColour = borderColour;
        }

        protected void RenderQuad()
        {
            CustomVertex.TransformedColored[] vertices = new CustomVertex.TransformedColored[4];

            vertices[0].Position = new Vector4((float)Rect.Left,   (float)Rect.Top, 0, 1.0f);
            vertices[1].Position = new Vector4((float)Rect.Right,  (float)Rect.Top, 0, 1.0f);
            vertices[2].Position = new Vector4((float)Rect.Left,   (float)Rect.Bottom, 0, 1.0f);
            vertices[3].Position = new Vector4((float)Rect.Right,  (float)Rect.Bottom, 0, 1.0f);

            vertices[0].Color = (Int32)Colour;
            vertices[1].Color = (Int32)Colour;
            vertices[2].Color = (Int32)Colour;
            vertices[3].Color = (Int32)Colour;

            Device.VertexFormat = CustomVertex.TransformedColored.Format;

            Device.DrawUserPrimitives(PrimitiveType.TriangleStrip, 2, vertices);

            if (BorderColour != 0x00000000)
            {
                vertices[0].Color = (Int32)BorderColour;
                vertices[1].Color = (Int32)BorderColour;
                vertices[2].Color = (Int32)BorderColour;
                vertices[3].Color = (Int32)BorderColour;

                UInt16[] indices = { 0, 1, 3, 2, 0 };
                Device.DrawIndexedUserPrimitives(PrimitiveType.LineStrip, 0, 4, 4, indices, true, vertices);
            }
        }

        public virtual void Render()
        {
            RenderQuad();
        }
    }
}