﻿using System;
using System.Drawing;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    class ModelExScrollBar : ModelExQuad
    {
        bool Vertical;

        UInt32 ControlColour;
        UInt32 ControlHighlightColour;

        ModelExQuad Plus;
        ModelExQuad Minus;
        ModelExQuad Slider;

        Point MinimumSliderPosition;
        Point MaximumSliderPosition;

        bool IsDraggingSlider;
        bool IsHoldingPlus;
        bool IsHoldingMinus;

        Point MouseDownPosition;
        Point MouseDownSliderPosition;

        Rectangle PrevRect;

        public int Position;
        public int Range;

        public ModelExScrollBar(Device device, Rectangle rect, UInt32 controlColour, UInt32 controlHighlightColour, UInt32 borderColour, UInt32 backColour, bool vertical, int range) :
            base(device, rect, backColour, borderColour)
        {
            Vertical = vertical;

            ControlColour = controlColour;
            ControlHighlightColour = controlHighlightColour;

            Plus = new ModelExQuad(device, rect, controlColour, borderColour);
            Minus = new ModelExQuad(device, rect, controlColour, borderColour);
            Slider = new ModelExQuad(device, rect, controlColour, borderColour);

            Range = range;

            UpdatePositions();
        }

        public bool HandleMouseUp(Point mousePosition)
        {
            IsDraggingSlider = false;
            IsHoldingPlus = false;
            IsHoldingMinus = false;

            Plus.Colour = ControlColour;
            Minus.Colour = ControlColour;
            Slider.Colour = ControlColour;

            return true;
        }

        public bool HandleMouseDown(Point mousePosition)
        {
            IsDraggingSlider = false;
            IsHoldingPlus = false;
            IsHoldingMinus = false;

            Plus.Colour = ControlColour;
            Minus.Colour = ControlColour;
            Slider.Colour = ControlColour;

            if (Slider.Rect.Contains(mousePosition))
            {
                IsDraggingSlider = true;
                MouseDownPosition = mousePosition;
                MouseDownSliderPosition = Slider.Rect.Location;
                Slider.Colour = ControlHighlightColour;
            }
            else if (Plus.Rect.Contains(mousePosition))
            {
                IsHoldingPlus = true;
                Plus.Colour = ControlHighlightColour;
            }
            else if (Minus.Rect.Contains(mousePosition))
            {
                IsHoldingMinus = true;
                Minus.Colour = ControlHighlightColour;
            }

            return true;
        }

        public bool HandleMouseMove(Point mousePosition)
        {
            if (IsDraggingSlider)
            {
                if (Vertical)
                {
                    Slider.Rect.Y = MouseDownSliderPosition.Y + (mousePosition.Y - MouseDownPosition.Y);
                    if (Slider.Rect.Y < MinimumSliderPosition.Y)
                    {
                        Slider.Rect.Y = MinimumSliderPosition.Y;
                    }
                    else if (Slider.Rect.Y > MaximumSliderPosition.Y)
                    {
                        Slider.Rect.Y = MaximumSliderPosition.Y;
                    }

                    int sliderPosition = (Slider.Rect.Y - MinimumSliderPosition.Y) * Range;
                    Position = sliderPosition / (MaximumSliderPosition.Y - MinimumSliderPosition.Y);
                }
                else
                {
                    Slider.Rect.X = MouseDownSliderPosition.X + (mousePosition.X - MouseDownPosition.X);
                    if (Slider.Rect.X < MinimumSliderPosition.X)
                    {
                        Slider.Rect.X = MinimumSliderPosition.X;
                    }
                    else if (Slider.Rect.X > MaximumSliderPosition.X)
                    {
                        Slider.Rect.X = MaximumSliderPosition.X;
                    }

                    int sliderPosition = (Slider.Rect.X - MinimumSliderPosition.X) * Range;
                    Position = sliderPosition / (MaximumSliderPosition.X - MinimumSliderPosition.X);
                }
            }

            return true;
        }

        void Update()
        {
            if (IsHoldingPlus)
            {
                Position++;
            }
            else if (IsHoldingMinus)
            {
                Position--;
            }
        }

        public void UpdatePositions()
        {
            if (Rect == PrevRect)
            {
                return;
            }

            Plus.Rect = Rect;
            Minus.Rect = Rect;
            Slider.Rect = Rect;

            if (Vertical)
            {
                Plus.Rect.Height = Plus.Rect.Width;
                Minus.Rect.Height = Minus.Rect.Width;
                Minus.Rect.Y = Rect.Bottom - Minus.Rect.Height;
                Slider.Rect.Height = Slider.Rect.Width;

                MinimumSliderPosition.Y = Plus.Rect.Bottom + 1;
                MaximumSliderPosition.Y = Minus.Rect.Top - Minus.Rect.Height - 1;
                Slider.Rect.Y = MinimumSliderPosition.Y;
            }
            else
            {
                Plus.Rect.Width = Plus.Rect.Height;
                Minus.Rect.Width = Minus.Rect.Height;
                Minus.Rect.X = Rect.Right - Minus.Rect.Width;
                Slider.Rect.Width = Slider.Rect.Height;

                MinimumSliderPosition.X = Plus.Rect.Right + 1;
                MaximumSliderPosition.X = Minus.Rect.Left - Minus.Rect.Width - 1;
                Slider.Rect.X = MinimumSliderPosition.X;
            }

            IsDraggingSlider = false;
            IsHoldingPlus = false;
            IsHoldingMinus = false;

            if (Vertical)
            {
                int sliderPosition = (Slider.Rect.Y - MinimumSliderPosition.Y) * Range;
                Position = sliderPosition / (MaximumSliderPosition.Y - MinimumSliderPosition.Y);
            }
            else
            {
                int sliderPosition = (Slider.Rect.X - MinimumSliderPosition.X) * Range;
                Position = sliderPosition / (MaximumSliderPosition.X - MinimumSliderPosition.X);
            }

            PrevRect = Rect;
        }

        public override void Render()
        {
            base.Render();

            Plus.Render();
            Minus.Render();
            Slider.Render();
        }
    }
}
