﻿using System;
using System.Collections.Generic;
using System.Drawing;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    class ModelExStatusBox : ModelExTextBox
    {
        public ModelExFont StatusFont;
        public String[] StatusText;
        public int Status;
        public Rectangle StatusTextRect;

        public ModelExStatusBox(Device device, Rectangle rect, UInt32 colour, UInt32 borderColour, UInt32 textColour, UInt32 textColourDisabled, ModelExFont font, Point margin, ModelExFont statusFont, Point statusMargin, String[] statusText) :
            base(device, rect, colour, borderColour, textColour, textColourDisabled, font, margin)
        {
            StatusFont = statusFont;
            StatusText = statusText;
            Status = 0;
            UpdatePositions();
        }

        public ModelExStatusBox(Device device, String text, Rectangle rect, UInt32 colour, UInt32 borderColour, UInt32 textColour, UInt32 textColourDisabled, ModelExFont font, Point margin, ModelExFont statusFont, Point statusMargin, String[] statusText) :
            base(device, text, rect, colour, borderColour, textColour, textColourDisabled, font, margin)
        {
            StatusFont = statusFont;
            StatusText = statusText;
            Status = 0;
            UpdatePositions();
        }

        public new void UpdatePositions()
        {
            Rectangle textRect = Rect;
            Rectangle statusTextRect = Rect;
            textRect.Inflate(-Margin.X, -Margin.Y);
            statusTextRect.Inflate(-Margin.X, -Margin.Y);
            statusTextRect.X = statusTextRect.Right - statusTextRect.Height;
            statusTextRect.Width = statusTextRect.Height;

            if (StatusText != null)
            {
                textRect.Width = textRect.Width - statusTextRect.Height;
            }

            TextRect = textRect;
            StatusTextRect = statusTextRect;
        }

        public override void Render()
        {
            RenderQuad();

            uint currentColour = TextColour;
            if (!Enabled)
            {
                currentColour = TextColourDisabled;
            }

            if (StatusText != null)
            {
                Font.Print(Text, TextRect, currentColour);
                StatusFont.Print(StatusText[Status], StatusTextRect, currentColour);
            }
            else
            {
                Font.Print(Text, TextRect, currentColour);
            }
        }
    }
}
