﻿using System;
using System.Drawing;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    class ModelExTextBox : ModelExQuad
    {
        public ModelExFont Font;
        public UInt32 TextColour = 0xFF000000;
        public UInt32 TextColourDisabled = 0xFF808080;
        public String Text;
        public Rectangle TextRect;
        public Point Margin;
        public Point Offset;

        public ModelExTextBox(Device device, Rectangle rect, UInt32 colour, UInt32 borderColour, UInt32 textColour, UInt32 textColourDisabled, ModelExFont font, Point margin) :
            base(device, rect, colour, borderColour)
        {
            TextColour = textColour;
            TextColourDisabled = textColourDisabled;
            Font = font;
            Margin = margin;
            Offset = new Point();
            UpdatePositions();
        }

        public ModelExTextBox(Device device, String text, Rectangle rect, UInt32 colour, UInt32 borderColour, UInt32 textColour, UInt32 textColourDisabled, ModelExFont font, Point margin) :
            base(device, rect, colour, borderColour)
        {
            Text = text;
            TextColour = textColour;
            TextColourDisabled = textColourDisabled;
            Font = font;
            Margin = margin;
            UpdatePositions();
        }

        public void UpdatePositions()
        {
            TextRect = Rect;
            TextRect.Inflate(-Margin.X, -Margin.Y);
        }

        public override void Render()
        {
            RenderQuad();

            Viewport backUpViewport = Device.Viewport;
            Viewport textViewport = new Viewport();

            textViewport.X = TextRect.X;
            textViewport.Width = TextRect.Width;
            textViewport.Y = TextRect.Y;
            textViewport.Height = TextRect.Height;

            Rectangle textRect = TextRect;
            textRect.Offset(Offset);
            textRect.Height = 2000;

            Device.Viewport = textViewport;
            if (Enabled)
            {
                Font.Print(Text, textRect, TextColour);
            }
            else
            {
                Font.Print(Text, textRect, TextColourDisabled);
            }
            Device.Viewport = backUpViewport;
        }
    }
}
