using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    class ArcBall
    {
        public delegate void MoveHandler();

        protected Control               m_xControl;
        protected event MoveHandler     m_xMoveHandler;
        protected Quaternion            m_xCurrentRotation;
        protected Quaternion            m_xStartRotation;
        protected Vector3               m_xStartVector;
        protected Vector3               m_xCurrentVector;
        protected bool                  m_bInversedControls;
        protected bool                  m_bMovingArcball;
        protected bool                  m_bBeginMove;

        public MoveHandler Move
        {
            get { return m_xMoveHandler; }
            set { m_xMoveHandler = value; }
        }
        public Quaternion Rotation
        {
            get { return m_xCurrentRotation; }
            set
            {
                m_xCurrentRotation = value;
                EndMove();
            }
        }
        public bool IsMoving
        {
            get { return m_bMovingArcball; }
        }
        public bool InversedControls
        {
            get { return m_bInversedControls; }
        }

        public ArcBall(ref Control xControl, bool bInversedControls)
        {
            m_xControl          = xControl;
            m_xMoveHandler      = null;
            m_xCurrentRotation  = Quaternion.Identity;
            m_xStartRotation    = Quaternion.Identity;
            m_xStartVector      = new Vector3();
            m_xCurrentVector    = new Vector3();
            m_bInversedControls = bInversedControls;
            m_bMovingArcball    = false;
            m_bBeginMove        = false;
        }

        public void BeginMove()
        {
            if (!m_bMovingArcball)
            {
                m_xStartRotation    = m_xCurrentRotation;
                m_bMovingArcball    = true;
                m_bBeginMove        = true;
            }
        }

        public void EndMove()
        {
            if (m_bMovingArcball)
            {
                m_xStartRotation    = m_xCurrentRotation;
                m_bMovingArcball  = false;
                m_bBeginMove        = false;
            }
        }

        public virtual void OnMove(int iX, int iY, Quaternion xReference)
        {
            if (!m_bMovingArcball)
            {
                return;
            }

            HandleMove(iX, iY, xReference);

            if (m_xMoveHandler != null)
            {
                m_xMoveHandler();
            }

            if (m_bBeginMove)
            {
                m_bBeginMove = false;
            }
        }

        protected virtual void HandleMove(int iX, int iY, Quaternion xReference)
        {
            if (m_bBeginMove)
            {
                m_xStartVector = ScreenToVector(iX, iY);
            }

            m_xCurrentVector = ScreenToVector(iX, iY);
            float fDotPart = Vector3.Dot(m_xStartVector, m_xCurrentVector);
            Vector3 xPart = Vector3.Cross(m_xStartVector, m_xCurrentVector);
            Quaternion xQuatFromBallPoints = new Quaternion(xPart.X, xPart.Y, -xPart.Z, fDotPart);

            if (m_bInversedControls)
            {
                m_xCurrentRotation = xQuatFromBallPoints * m_xStartRotation;
            }
            else
            {
                m_xCurrentRotation = m_xStartRotation;
                m_xCurrentRotation *= Quaternion.Invert(xReference);
                m_xCurrentRotation *= Quaternion.Invert(xQuatFromBallPoints);
                m_xCurrentRotation *= xReference;
            }

            m_xCurrentRotation.Normalize();
        }

        protected Vector3 ScreenToVector(int iX, int iY)
        {
            Vector3 xScreenCentre = new Vector3(
                m_xControl.Width / 2.0f,
                m_xControl.Height / 2.0f,
                0.0f
            );

            int iScreenDiameter = Math.Min(m_xControl.Width, m_xControl.Height);

            Vector3 xResult = new Vector3(
                (iX - xScreenCentre.X) / (iScreenDiameter / 2.0f),
                (xScreenCentre.Y - iY) / (iScreenDiameter / 2.0f),
                0.0f
            );

            float fMagnitude = xResult.X * xResult.X + xResult.Y * xResult.Y;

            if (fMagnitude > 1.0)
            {
                float fScale = 1.0f / (float)Math.Sqrt((double)fMagnitude);
                xResult *= fScale;
            }
            else
            {
                xResult.Z = (float)Math.Sqrt((double)(1.0f - fMagnitude));
            }

            return xResult;
        }
    }
}
