texture xTexture0;
bool bLighting;
bool bColorVertex;
int iCullMode;
int iFillMode;
int iColorOperation;
int iAlphaOperation;
int iTextureFilter;
int iAmbientMaterialSource;
int iDiffuseMaterialSource;

float4x4 xWorld : WORLD;
float4x4 xView : VIEW;
float4x4 xWorldViewProj : WORLDVIEWPROJ;

float4 xMaterialAmbient;
float4 xMaterialDiffuse;
float4 xMaterialSpecular;
float4 xMaterialEmissive;

struct Light
{
    float4 Position;
	float4 Direction;
    float4 Diffuse;
    float4 Specular;
    float4 Ambient;
    float4 Attenuation;
};

Light xLights[1];

//application to vertex structure
struct a2v
{
   float4 position : POSITION0;
   float3 normal : NORMAL;
   float4 diffuse : COLOR0;
   float2 tex0 : TEXCOORD0;
};

//vertex to pixel processing structure
struct v2p
{
   float4 position : POSITION0;
   float4 color0 : COLOR0;
   float4 color1 : COLOR1;
   float2 tex0 : TEXCOORD0;
};

void vs( in a2v IN, out v2p OUT ) 
{
    OUT.position = mul(IN.position, xWorldViewProj);

	float3 xWorldNormal = normalize(mul(IN.normal, xWorld));
	float3 xWorldPosition = mul(IN.position, xWorld);
	float3 xCameraPosition = mul(xWorldPosition, xView);

	if (bLighting)
	{
		float4 diffuse = float4(0,0,0,1);
		float4 specular = float4(0,0,0,1);

	    for (int i = 0; i < 1; i++)
	    {
	        float3 toLight = xLights[i].Position.xyz - xWorldPosition;
	        float lightDist = length(toLight);
	        float fAttenuation = 1.0 / dot(xLights[i].Attenuation, float4(1, lightDist, lightDist * lightDist, 0));
	        float3 lightDir = normalize(xLights[i].Direction); // normalize(toLight);
	        float3 halfAngle = normalize(normalize(-xCameraPosition) + lightDir);

	        diffuse += max(0, dot(lightDir, xWorldNormal) * xLights[i].Diffuse * fAttenuation) + xLights[i].Ambient;
	        specular += max(0, pow(dot(halfAngle, xWorldNormal), 64) * xLights[i].Specular * fAttenuation);
	    }
	
		OUT.color0 = diffuse;
		OUT.color1 = specular;
	}
	else
	{
		OUT.color0 = IN.diffuse;
		OUT.color1 = float4(0, 0, 0, 1);
	}

	OUT.tex0 = IN.tex0;
}

float4 ps() : COLOR
{
     return float4(1,1,1,1);
}

technique FixedFunctionPipeline
{
	pass p0
	{
		Texture[0] = <xTexture0>;
		MaterialAmbient = <xMaterialAmbient>;
		MaterialDiffuse = <xMaterialDiffuse>;
		Lighting = <bLighting>;
		ColorVertex = <bColorVertex>;
		CullMode = <iCullMode>;
		FillMode = <iFillMode>;
		ZEnable = TRUE;
		ZWriteEnable = TRUE;
		DitherEnable = TRUE;
		SpecularEnable = FALSE; // Use Color1 if true.
		NormalizeNormals = TRUE;
		AlphaTestEnable = TRUE;
		AlphaFunc = GREATER;
		AlphaRef = 0xA0;
		ColorOp[0] = <iColorOperation>;
		ColorArg1[0] = TEXTURE;
		ColorArg2[0] = CURRENT;
		AlphaOp[0] = <iAlphaOperation>;
		AlphaArg1[0] = TEXTURE;
		AlphaArg2[0] = DIFFUSE;
		MagFilter[0] = <iTextureFilter>;
		MinFilter[0] = <iTextureFilter>;
		AmbientMaterialSource = <iAmbientMaterialSource>;
		DiffuseMaterialSource = <iDiffuseMaterialSource>;
		WorldTransform[0] = <xWorld>;
	}
}

technique Shader
{
    pass p0
    {
        vertexshader = compile vs_1_1 vs();
		//pixelshader = compile ps_1_1 ps();
		
		Texture[0] = <xTexture0>;
		MaterialAmbient = <xMaterialAmbient>;
		MaterialDiffuse = <xMaterialDiffuse>;
		Lighting = <bLighting>;
		ColorVertex = <bColorVertex>;
		CullMode = <iCullMode>;
		FillMode = <iFillMode>;
		ZEnable = TRUE;
		ZWriteEnable = TRUE;
		DitherEnable = TRUE;
		SpecularEnable = FALSE; // Use Color1 if true.
		NormalizeNormals = TRUE;
		AlphaTestEnable = TRUE;
		AlphaFunc = GREATER;
		AlphaRef = 0xA0;
		ColorOp[0] = <iColorOperation>;
		ColorArg1[0] = TEXTURE;
		ColorArg2[0] = CURRENT;
		AlphaOp[0] = <iAlphaOperation>;
		AlphaArg1[0] = TEXTURE;
		AlphaArg2[0] = DIFFUSE;
		MagFilter[0] = <iTextureFilter>;
		MinFilter[0] = <iTextureFilter>;
		AmbientMaterialSource = <iAmbientMaterialSource>;
		DiffuseMaterialSource = <iDiffuseMaterialSource>;
		WorldTransform[0] = <xWorld>;
    }
}